@@ -15,9 +15,16 @@ import json
 import uuid
 from hashlib import blake2s
 
-import yaml
+from typing import Dict, Any, Sequence, Tuple
+
+from skytools.basetypes import Cursor
 import skytools
+import yaml
+
+from pgq.event import Event
 from londiste.handler import TableHandler
+import londiste.util
+
 
 __all__ = ['Obfuscator']
 
@@ -31,6 +38,7 @@ HASH64 = 'hash64'
 HASH128 = 'hash'
 SKIP = 'skip'
 
+
 def as_bytes(data):
     """Convert input string or json value into bytes.
     """
@@ -48,6 +56,7 @@ def as_bytes(data):
     # no point hashing str() of list or dict
     raise ValueError('Invalid input type for hashing: %s' % type(data))
 
+
 def hash32(data):
     """Returns hash as 32-bit signed int.
     """
@@ -56,6 +65,7 @@ def hash32(data):
     hash_bytes = blake2s(as_bytes(data), digest_size=4, key=_KEY).digest()
     return int.from_bytes(hash_bytes, byteorder='big', signed=True)
 
+
 def hash64(data):
     """Returns hash as 64-bit signed int.
     """
@@ -64,6 +74,7 @@ def hash64(data):
     hash_bytes = blake2s(as_bytes(data), digest_size=8, key=_KEY).digest()
     return int.from_bytes(hash_bytes, byteorder='big', signed=True)
 
+
 def hash128(data):
     """Returns hash as 128-bit variant 0 uuid.
     """
@@ -79,6 +90,26 @@ def hash128(data):
 
     return str(uuid.UUID(int=hash_int))
 
+
+def data_to_dict(data, column_list):
+    """Convert data received from copy to dict
+    """
+    if data[-1] == '\n':
+        data = data[:-1]
+
+    vals = [skytools.unescape_copy(value) for value in data.split('\t')]
+    row = dict(zip(column_list, vals))
+    return row
+
+
+def obf_vals_to_data(obf_vals):
+    """Converts obfuscated values back to copy data
+    """
+    vals = [skytools.quote_copy(value) for value in obf_vals]
+    obf_data = '\t'.join(vals) + '\n'
+    return obf_data
+
+
 def obf_json(json_data, rule_data):
     """JSON cleanup.
 
@@ -119,11 +150,12 @@ def obf_json(json_data, rule_data):
         return hash128(json_data)
     raise ValueError('Invalid rule value: %r' % rule_data)
 
+
 class Obfuscator(TableHandler):
     """Default Londiste handler, inserts events into tables with plain SQL.
     """
     handler_name = 'obfuscate'
-    obf_map = {}
+    obf_map: Dict[str, Any] = {}
 
     @classmethod
     def load_conf(cls, cf):
@@ -145,7 +177,7 @@ class Obfuscator(TableHandler):
 
         Returns either string (sql event) or dict (urlenc event).
         """
-        row = super(Obfuscator, self).parse_row_data(ev)
+        row = super().parse_row_data(ev)
 
         rule_data = self._get_map(self.table_name, row)
         dst = {}
@@ -186,24 +218,16 @@ class Obfuscator(TableHandler):
     def obf_copy_row(self, data, column_list, src_tablename):
         """Apply obfuscation to one row
         """
-        if data[-1] == '\n':
-            data = data[:-1]
-        else:
-            self.log.warning('Unexpected line from copy without end of line.')
-
-        vals = data.split('\t')
-
-        row = {name: value for name, value in zip(column_list, vals)}
+        row = data_to_dict(data, column_list)
         obf_col_map = self._get_map(src_tablename, row)
 
         obf_vals = []
-        for field, value in zip(column_list, vals):
+        for field, value in row.items():
             action = obf_col_map.get(field, SKIP)
 
             if isinstance(action, dict):
-                str_val = skytools.unescape_copy(value)
-                obf_val = self.obf_json(str_val, action)
-                obf_vals.append(skytools.quote_copy(obf_val))
+                obf_val = self.obf_json(value, action)
+                obf_vals.append(obf_val)
                 continue
             elif action == KEEP:
                 obf_vals.append(value)
@@ -211,24 +235,24 @@ class Obfuscator(TableHandler):
             elif action == SKIP:
                 continue
 
-            str_val = skytools.unescape_copy(value)
-            if str_val is None:
+            if value is None:
                 obf_vals.append(value)
             elif action == BOOL:
-                obf_val = str(bool(str_val) and 't' or 'f')
+                obf_val = str(bool(value) and 't' or 'f')
                 obf_vals.append(obf_val)
             elif action == HASH32:
-                obf_val = str(hash32(str_val))
+                obf_val = str(hash32(value))
                 obf_vals.append(obf_val)
             elif action == HASH64:
-                obf_val = str(hash64(str_val))
+                obf_val = str(hash64(value))
                 obf_vals.append(obf_val)
             elif action == HASH128:
-                obf_val = hash128(str_val)
+                obf_val = hash128(value)
                 obf_vals.append(obf_val)
             else:
                 raise ValueError('Invalid value for action: %s' % action)
-        obf_data = '\t'.join(obf_vals) + '\n'
+
+        obf_data = obf_vals_to_data(obf_vals)
         return obf_data
 
     def real_copy(self, src_tablename, src_curs, dst_curs, column_list):
@@ -252,8 +276,61 @@ class Obfuscator(TableHandler):
                                   dst_tablename=self.dest_table,
                                   write_hook=_write_hook)
 
+    def real_copy_threaded(
+        self,
+        src_real_table: str,
+        src_curs: Cursor,
+        dst_db_connstr: str,
+        column_list: Sequence[str],
+        config_file: str,
+        config_section: str,
+        parallel: int = 1,
+    ) -> Tuple[int, int]:
+        with skytools.connect_database(dst_db_connstr) as dst_db:
+            with dst_db.cursor() as dst_curs:
+                condition = self.get_copy_condition(src_curs, dst_curs)
+            dst_db.commit()
+
+        obf_col_map = self._get_map(src_real_table)
+
+        new_list = []
+        for col in column_list:
+            action = obf_col_map.get(col, SKIP)
+            if action != SKIP:
+                new_list.append(col)
+        column_list = new_list
+
+        def _write_hook(_, data):
+            return self.obf_copy_row(data, column_list, src_real_table)
+
+        return londiste.util.full_copy_parallel(
+            src_real_table, src_curs,
+            dst_db_connstr=dst_db_connstr,
+            dst_tablename=self.dest_table,
+            condition=condition,
+            column_list=column_list,
+            write_hook=_write_hook,
+            parallel=parallel,
+        )
+
+    def get_copy_event(self, ev, queue_name):
+        row = self.parse_row_data(ev)
+
+        if len(ev.type) == 1:
+            ev_data = row
+        elif ev.data[0] == '{':
+            ev_data = skytools.json_encode(row)
+        else:
+            ev_data = skytools.db_urlencode(row)
+
+        ev_row = ev._event_row.copy()
+        ev_row['ev_data'] = ev_data
+        return Event(queue_name, ev_row)
+
+
 __londiste_handlers__ = [Obfuscator]
 
 if __name__ == '__main__':
     import doctest
     doctest.testmod()
+
